#!/usr/bin/env python3

import time
import warnings
import numpy
import astropy.constants

from astropy.io import fits

from captools import ppxf_util

from mangadap.drpfits import DRPFits
from mangadap.proc.reductionassessments import ReductionAssessment
from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectra
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel
from mangadap.proc.ppxffit import PPXFFit
from mangadap.util.instrument import spectrum_velocity_scale
from mangadap.util.fitsutil import DAPFitsUtil
from mangadap.util.fileio import init_record_array
from mangadap.proc.emissionlinemoments import EmissionLineMoments
from mangadap.proc.spectralfitting import EmissionLineFit
from mangadap.proc.emissionlinemodel import EmissionLineModelDef, EmissionLineModel
from mangadap.par.emissionlinedb import EmissionLineDB
from mangadap.proc.spectralindices import SpectralIndices
from mangadap.dapfits import construct_maps_file, construct_cube_file

from mangadap.contrib.xjmc import emline_fitter_with_ppxf

#-----------------------------------------------------------------------------

class XJMCEmissionLineFitter(EmissionLineFit):
    def __init__(self, par=None):
        if par is None:
            # Set the default parameter set.  The guess_redshift,
            # stellar_continuum, and emission_lines values can be filled
            # by EmissionLineModel._fill_method_par()
            par = { 'guess_redshift': None,     # The guess redshift for each binned spectrum
                    'stellar_continuum': None,  # The StellarContinuumModel object
                    'emission_lines': None,     # The EmissionLineDB object
                    'degree': -1,                # Additive polynomial order
                    'mdegree': 10 }              # Multiplicative polynomial order
        EmissionLineFit.__init__(self, 'XJMC', None, par=par)


    def line_database(self):
        """
        For use in the output to distinguish the lines being fit.
        """
        neml = 11
        emission_line_name = [ 'Hd', 'Hg', 'Hb', 'Ha', 'OII', 'OII', 'SII', 'SII', 'OIIId', 'OId',
                               'NIId']
        emission_line_restwave = [ 4103, 4342, 4861, 6564, 3727, 3729, 6718, 6732, 4984, 6334,
                                   6567 ]

        name_len = numpy.amax([ len(n) for n in emission_line_name ])
        db = init_record_array(neml, [ ('NAME','<U{0:d}'.format(name_len)), ('RESTWAVE', float) ])
        db['NAME'] = emission_line_name
        db['RESTWAVE'] = emission_line_restwave
        return db
        
    
    def fit(self, binned_spectra, par=None, loggers=None, quiet=False):
        if par is not None:
            self.par = par

        # Check the parameter keys
        required_keys = [ 'guess_redshift', 'stellar_continuum', 'emission_lines', 'degree',
                          'mdegree' ]
        if numpy.any([ reqk not in self.par.keys() for reqk in required_keys ]):
            raise ValueError('Parameter dictionary does not have all the required keys.')

        # Wavelengths are in vacuum
        wave0 = binned_spectra['WAVE'].data.copy()
        # Velocity step per pixel
        velscale = spectrum_velocity_scale(wave0)
        
        # Get the best-fitting stellar kinematics for the binned spectra
        # And correct sigmas with instrumental resolutions (if convolved templates are to be used)
        if self.par['stellar_continuum'] is None \
                or not isinstance(self.par['stellar_continuum'], StellarContinuumModel):
            raise ValueError('Must provide StellarContinuumModel object as the '
                             '\'stellar_continuum\' item in the parameter dictionary')

        stars_vel, stars_sig = self.par['stellar_continuum'].matched_guess_kinematics(
                                            binned_spectra, cz=True, corrected=True, nearest=True)

        # Convert the input stellar velocity from redshift (c*z) to ppxf velocity (c*log(1+z))
        stars_vel = PPXFFit.revert_velocity(stars_vel,0)[0]
        
        # Get the stellar templates and the template resolution;
        # shape is (Nstartpl, Ntplwave)
        stars_templates = self.par['stellar_continuum'].get_template_library(
                            velocity_offset=numpy.median(stars_vel), match_to_drp_resolution=True)
        stars_templates_wave = stars_templates['WAVE'].data.copy()
        template_sres = stars_templates['SPECRES'].data[0,:]
        stars_templates = stars_templates['FLUX'].data.copy()            
        velscale_ratio = self.par['stellar_continuum'].method['fitpar']['velscale_ratio']

        # Set mask for the galaxy spectra according to the templates wave range
        mask = PPXFFit.fitting_mask(tpl_wave=stars_templates_wave, obj_wave=wave0,
                                    velscale=velscale, velscale_ratio=velscale_ratio,
                                    velocity_offset=numpy.median(stars_vel))[0]
        wave = wave0[mask]

        # Calculate the velocity offset between the masked spectra and the tempaltes
        dv = -PPXFFit.ppxf_tpl_obj_voff(stars_templates_wave, wave, velscale,
                                            velscale_ratio=velscale_ratio)
        
        # UNBINNED DATA:
        # Flux and noise masked arrays; shape is (Nspaxel,Nwave) where
        # Nspaxel is Nx*Ny
        # Bin ID from VOR10 reference file used to mask buffer spaxels
        #binid0 = binned_spectra['BINID'].data
        # Create mask for parsing the input data
        #binid = binid0.reshape(-1)
        #mask_spaxel = ~(binid==-1)
        # pPXF would run into problems if dircetly masked arrays were used
        # So both fluxes and their masks are to be provided as input

#        flux00 = binned_spectra.drpf.copy_to_masked_array(flag=['DONOTUSE', 'FORESTAR'])
#        mask_drp0 = ~flux00.mask
#        flux = binned_spectra.drpf.copy_to_array(ext='FLUX')
#        ivar = binned_spectra.drpf.copy_to_array(ext='IVAR')
#        flux0, ivar0 = binned_spectra.galext.apply(flux, ivar=ivar, deredden=True)
#        noise0 = numpy.power(ivar0.data, -0.5)
#        mask_drp = mask_drp0.reshape(-1, mask_drp0.shape[-1])[:,mask]
#        flux = flux0.data[:,mask]
#        noise = noise0[:,mask]

        flux0 = binned_spectra.drpf.copy_to_masked_array(flag=['DONOTUSE', 'FORESTAR'])
        ivar0 = binned_spectra.drpf.copy_to_masked_array(ext='IVAR', flag=['DONOTUSE', 'FORESTAR'])
        flux0, ivar0 = binned_spectra.galext.apply(flux0, ivar=ivar0, deredden=True)
        noise = numpy.ma.power(ivar0, -0.5)
        noise[numpy.invert(noise > 0)] = numpy.ma.masked

        mask_drp = numpy.invert(flux0.mask | noise.mask)[:,mask]
        flux = flux0.data[:,mask]
        noise = noise.filled(0.0)[:,mask]

        # stack_sres sets whether or not the spectral resolution is
        # determined on a per-spaxel basis or with a single vector
        sres = binned_spectra.drpf.spectral_resolution(toarray=True, fill=True) \
                    if binned_spectra.method['stackpar']['stack_sres'] else \
                    binned_spectra.drpf.spectral_resolution(ext='SPECRES', toarray=True, fill=True)
        sres = sres[:,mask]

        # Spaxel coordinates; shape is (Nspaxel,)
        x = binned_spectra.rdxqa['SPECTRUM'].data['SKY_COO'][:,0]
        y = binned_spectra.rdxqa['SPECTRUM'].data['SKY_COO'][:,1]

        # BINNED DATA:
        # Binned flux and binned noise masked arrays; shape is (Nbin,Nwave)

#        flux_binned00 = binned_spectra.copy_to_masked_array(flag=binned_spectra.do_not_fit_flags())
#        mask_binned0 = ~flux_binned00.mask
#        flux_binned0 = binned_spectra.copy_to_array(ext='FLUX')
#        noise_binned0 = numpy.power(binned_spectra.copy_to_array(ext='IVAR'), -0.5)
#        mask_binned = mask_binned0[:,mask]
#        flux_binned = flux_binned0[:,mask]
#        noise_binned = noise_binned0[:,mask]
#        sres_binned0 = binned_spectra.copy_to_array(ext='SPECRES')
#        sres_binned = sres_binned0[:,mask]

        flux_binned = binned_spectra.copy_to_masked_array(flag=binned_spectra.do_not_fit_flags())
        noise_binned = numpy.ma.power(binned_spectra.copy_to_masked_array(ext='IVAR',
                                            flag=binned_spectra.do_not_fit_flags()), -0.5)
        noise_binned[numpy.invert(noise_binned > 0)] = numpy.ma.masked
        mask_binned = numpy.invert(flux_binned.mask | noise_binned.mask)[:,mask]
        flux_binned = flux_binned.data[:,mask]
        noise_binned = noise_binned.filled(0.0)[:,mask]
        sres_binned = binned_spectra.copy_to_array(ext='SPECRES')[:,mask]

        # Bin coordinates; shape is (Nbin,)
        x_binned = binned_spectra['BINS'].data['SKY_COO'][:,0]
        y_binned = binned_spectra['BINS'].data['SKY_COO'][:,1]
        
        # Set initial guesses for the velocity and velocity dispersion
        if self.par['guess_redshift'] is not None:
            # Use guess_redshift if provided
            guess_vel0 = self.par['guess_redshift'] * astropy.constants.c.to('km/s').value
            guess_vel = PPXFFit.revert_velocity(guess_vel0,0)[0]
            # And set default velocity dispersion to 100 km/s
            guess_sig = numpy.full(guess_vel.size, 100, dtype=float)
        elif self.par['stellar_continuum'] is not None:
            # Otherwise use the stellar-continuum result
            guess_vel, guess_sig = stars_vel.copy(), stars_sig.copy()
        else:
            raise ValueError('Cannot set guess kinematics; must provide either \'guess_redshift\' '
                             'or \'stellar_continuum\' in input parameter dictionary.')
        
        # Construct gas templates; shape is (Ngastpl, Ntplwave).
        # Template resolution matched between the stellar and gas
        # templates?
        # Decide whether to use convolved gas templates
        # Set the wavelength of lines to be in vacuum
        FWHM = wave/numpy.max(sres, axis=0)
        FWHM_binned = wave/sres_binned[0,:]
        def fwhm_drp(wave_len0):
            wave_len = wave_len0*(1 + numpy.median(self.par['guess_redshift']))
            index = numpy.argmin(abs(wave-wave_len[:,None]),axis=1) \
                        if numpy.asarray(wave_len) is wave_len \
                        else numpy.argmin(abs(wave-wave_len))
            return FWHM[index]
        def fwhm_binned(wave_len0):
            wave_len = wave_len0*(1 + numpy.median(self.par['guess_redshift']))
            index = numpy.argmin(abs(wave-wave_len[:,None]),axis=1) \
                        if numpy.asarray(wave_len) is wave_len \
                        else numpy.argmin(abs(wave-wave_len))
            return FWHM_binned[index]
        lam_range_gal = numpy.array([numpy.min(wave), numpy.max(wave)]) \
                            / (1 + numpy.median(self.par['guess_redshift']))
        gas_templates, gas_names, gas_wave = \
            ppxf_util.emission_lines(numpy.log(stars_templates_wave), lam_range_gal, fwhm_drp)
        gas_templates_binned, gas_names, gas_wave = \
            ppxf_util.emission_lines(numpy.log(stars_templates_wave), lam_range_gal, fwhm_binned)

        # Default polynomial orders
        degree = -1 if self.par['degree'] is None else self.par['degree']
        mdegree = 10 if self.par['mdegree'] is None else self.par['mdegree']
     
        # --------------------------------------------------------------
        # CALL TO EMLINE_FITTER_WITH_PPXF:
        # Input is:
        #   - wave: wavelength vector; shape is (Nwave,)
        #   - flux: observed, unbinned flux; masked array with shape
        #     (Nspaxel,Nwave)
        #   - noise: error in observed, unbinned flux; masked array with
        #     shape (Nspaxel,Nwave)
        #   - sres: spectral resolution (R=lambda/delta lambda) as a
        #     function of wavelength for each unbinned spectrum; shape
        #     is (Nspaxel,Nwave)
        #   - flux_binned: binned flux; masked array with shape
        #     (Nbin,Nwave)
        #   - noise_binned: noise in binned flux; masked array with
        #     shape (Nbin,Nwave)
        #   - sres_binned: spectral resolution (R=lambda/delta lambda)
        #     as a function of wavelength for each binned spectrum;
        #     shape is (Nbin,Nwave)
        #   - velscale: Velocity step per pixel
        #   - velscale_ratio: Ratio of velocity step per pixel in the
        #     observed data versus in the template data
        #   - dv: Velocity offset between the galaxy and template data
        #     due to the difference in the initial wavelength of the
        #     spectra
        #   - stars_vel: Velocity of the stellar component; shape is
        #     (Nbins,)
        #   - stars_sig: Velocity dispersion of the stellar component;
        #     shape is (Nbins,)
        #   - stars_templates: Stellar templates; shape is (Nstartpl,
        #     Ntplwave)
        #   - guess_vel: Initial guess velocity for the gas components;
        #     shape is (Nbins,)
        #   - guess_sig: Initial guess velocity dispersion for the gas
        #     components; shape is (Nbins,)
        #   - gas_templates: Gas template flux; shape is (Ngastpl,
        #     Ntplwave)
        #   - gas_names: Name of the gas templats; shape is (Ngastpl,)
        #   - template_sres: spectral resolution (R=lambda/delta
        #     lambda) as a function of wavelength for all the templates
        #     templates; shape is (Ntplwave,)
        #   - degree: Additive polynomial order
        #   - mdegree: Multiplicative polynomial order
        #   - x: On-sky spaxel x coordinates; shape is (Nspaxel,)
        #   - y: On-sky spaxel y coordinates; shape is (Nspaxel,)
        #   - x_binned: On-sky bin x coordinate; shape is (Nbin,)
        #   - y_binned: On-sky bin y coordinate; shape is (Nbin,)
        model_flux0, model_eml_flux0, model_mask0, model_binid, eml_flux, eml_fluxerr, \
                eml_kin, eml_kinerr, eml_sigmacorr \
                        = emline_fitter_with_ppxf(wave, flux, noise, sres, flux_binned,
                                                  noise_binned, velscale, velscale_ratio, dv,
                                                  stars_vel, stars_sig, stars_templates,
                                                  guess_vel, guess_sig, gas_templates,
                                                  gas_templates_binned, gas_names, template_sres,
                                                  degree, mdegree, x, y, x_binned, y_binned,
                                                  mask_binned, mask_drp,
                                                  numpy.median(self.par['guess_redshift']))
                                                  #, debug=True)
        # Output is:
        #   - model_flux: stellar-continuum + emission-line model; shape
        #     is (Nmod, Nwave); first axis is ordered by model ID number
        #   - model_eml_flux: model emission-line flux only; shape is
        #     (Nmod, Nwave); first axis is ordered by model ID number
        #   - model_mask: boolean or bit mask for fitted models; shape
        #     is (Nmod, Nwave); first axis is ordered by model ID number
        #   - model_binid: ID numbers assigned to each spaxel with a
        #     fitted model; any spaxel without a model should have
        #     model_binid = -1; the number of >-1 IDs must be Nmod;
        #     shape is (Nx,Ny) which is equivalent to:
        #       flux[:,0].reshape((numpy.sqrt(Nspaxel).astype(int),)*2).shape
        #   - eml_flux: Flux of each emission line; shape is (Nmod,Neml)
        #   - eml_fluxerr: Error in emission-line fluxes; shape is
        #     (Nmod, Neml)
        #   - eml_kin: Kinematics (velocity and velocity dispersion) of
        #     each emission line; shape is (Nmod,Neml,Nkin)
        #   - eml_kinerr: Error in the kinematics of each emission line
        #   - eml_sigmacorr: Quadrature corrections required to obtain
        #     the astrophysical velocity dispersion; shape is
        #     (Nmod,Neml); corrections are expected to be applied as
        #     follows:
        #       sigma = numpy.ma.sqrt( numpy.square(eml_kin[:,:,1])
        #                               - numpy.square(eml_sigmacorr))
        # --------------------------------------------------------------
        
        # Convert the output velocity back from ppxf velocity (c*log(1+z)) to redshift (c*z)
        eml_kin[:,:,0] = PPXFFit.convert_velocity(eml_kin[:,:,0],0)[0]
        
        # Mask output data according to model_binid
        model_binid = numpy.asarray(model_binid, dtype=numpy.int16)
        mask_id = (model_binid > -1)
        model_flux0 = model_flux0[mask_id]
        model_eml_flux0 = model_eml_flux0[mask_id]
        model_mask0 = model_mask0[mask_id]
        eml_flux = eml_flux[mask_id]
        eml_fluxerr = eml_fluxerr[mask_id]
        eml_kin = eml_kin[mask_id]
        eml_kinerr = eml_kinerr[mask_id]
        eml_sigmacorr = eml_sigmacorr[mask_id]
        
        #cube_binid = numpy.full_like(mask_spaxel, -1, dtype=numpy.int16)
        #cube_binid[mask_spaxel] = model_binid
        Nspaxel = x.shape[0]
        model_binid = model_binid.reshape((numpy.sqrt(Nspaxel).astype(int),)*2)
        # The ordered indices in the flatted bin ID map with/for each model
        model_srt = numpy.argsort(model_binid.ravel())[model_binid.ravel() > -1]

        # Construct the output emission-line database.  The data type
        # defined by EmissionLineFit._per_emission_line_dtype(); shape
        # is (Nmod,); parameters must be ordered by model ID number
        nmod = len(model_srt)
        neml = eml_flux.shape[1]
        nkin = eml_kin.shape[-1]
        model_eml_par = init_record_array(nmod,
                                EmissionLineFit._per_emission_line_dtype(neml, nkin, numpy.int16))
        model_eml_par['BINID'] = model_binid.ravel()[model_srt]
        model_eml_par['BINID_INDEX'] = numpy.arange(nmod)
        model_eml_par['MASK'][:,:] = 0
        model_eml_par['FLUX'] = eml_flux
        model_eml_par['FLUXERR'] = eml_fluxerr
        model_eml_par['KIN'] = eml_kin
        model_eml_par['KINERR'] = eml_kinerr
        model_eml_par['SIGMACORR'] = eml_sigmacorr

        # Include the equivalent width measurements
        if self.par['emission_lines'] is not None:
            EmissionLineFit.measure_equivalent_width(wave, flux[model_srt,:],
                                                     par['emission_lines'], model_eml_par)
        
        # Change back the wavelength range of models to match that of the galaxy's
        model_flux = numpy.zeros(flux0[mask_id].shape)
        model_eml_flux = numpy.zeros(flux0[mask_id].shape)
        model_mask = numpy.full_like(flux0[mask_id], 0, dtype=bool)
        model_flux[:,mask] = model_flux0
        model_eml_flux[:,mask] = model_eml_flux0
        model_mask[:,mask] = model_mask0
        
        # Calculate the "emission-line baseline" as the difference
        # between the stellar continuum model determined for the
        # kinematics and the one determined by the optimized
        # stellar-continuum + emission-line fit:
        if self.par['stellar_continuum'] is not None:
            # Construct the full 3D cube for the stellar continuum
            # models
            sc_model_flux, sc_model_mask \
                    = DAPFitsUtil.reconstruct_cube(binned_spectra.drpf.shape,
                                            self.par['stellar_continuum']['BINID'].data.ravel(),
                                            [ self.par['stellar_continuum']['FLUX'].data,
                                              self.par['stellar_continuum']['MASK'].data ])
            # Set any masked pixels to 0
            sc_model_flux[sc_model_mask>0] = 0.0

            # Construct the full 3D cube of the new stellar continuum
            # from the combined stellar-continuum + emission-line fit
            el_continuum = DAPFitsUtil.reconstruct_cube(binned_spectra.drpf.shape,
                                                        model_binid.ravel(),
                                                        model_flux - model_eml_flux)
            # Get the difference, restructure it to match the shape
            # of the emission-line models, and zero any masked pixels
            model_eml_base = (el_continuum - sc_model_flux).reshape(-1,wave0.size)[model_srt,:]
            if model_mask is not None:
                model_eml_base[model_mask==0] = 0.0
        else:
            model_eml_base = numpy.zeros(model_flux.shape, dtype=float)

        # Returned arrays are:
        #   - model_eml_flux: model emission-line flux only; shape is
        #     (Nmod, Nwave); first axis is ordered by model ID number
        #   - model_eml_base: difference between the combined fit and the
        #     stars-only fit; shape is (Nmod, Nwave); first axis is
        #     ordered by model ID number
        #   - model_mask: boolean or bit mask for fitted models; shape
        #     is (Nmod, Nwave); first axis is ordered by model ID number
        #   - model_fit_par: This provides the results of each fit;
        #     TODO: The is set to None.  Provide metrics of the ppxf fit
        #     to each spectrum?
        #   - model_eml_par: output model parameters; data type must be
        #     EmissionLineFit._per_emission_line_dtype(); shape is
        #     (Nmod,); parameters must be ordered by model ID number
        #   - model_binid: ID numbers assigned to each spaxel with a
        #     fitted model; any spaxel with a model should have
        #     model_binid = -1; the number of >-1 IDs must be Nmod;
        #     shape is (Nx,Ny)
        return model_eml_flux, model_eml_base, model_mask, None, model_eml_par, model_binid

    
#-----------------------------------------------------------------------------
if __name__ == '__main__':
    t = time.clock()

    # Set the plate, ifu, and initial velocity/redshift
    plate = 7815
    ifu = 6101
    # KBW testing edits
    vel = 5144.7681
    nsa_redshift = vel/astropy.constants.c.to('km/s').value
#    hdu_cat = fits.open('/Users/mac/Downloads/Monty_Python/Catalog/mpl5_cat.fits')
#    mask_z = [i['PLATEIFU'] == str(plate)+'-'+str(ifu) for i in hdu_cat[1].data]
#    nsa_redshift = hdu_cat[1].data[mask_z][0]['NSA_Z']
#    vel = nsa_redshift*astropy.constants.c.to('km/s').value
    
    # KBW testing edits
#    redux_path = '/Volumes/repo/MaNGA/redux/v2_2_0'
#    analysis_path = '/Users/westfall/Work/MaNGA/testing/mpl6/emissionlines/analysis/v2_2_0/test'
#    redux_path='/Users/mac/Desktop/DRP_MPL5/'
    redux_path='/Users/mac/Desktop/DRP_220/'    # v 2_2_0 redux path
    analysis_path='/Users/mac/Volumes/External/MaNGA/DAP_Output/MPL-5/2.0.2/'
    clobber = False
    
    # Read the DRP LOGCUBE file
    print('Reading DRP fits file.')
    drpf = DRPFits(plate, ifu, 'CUBE', read=True, redux_path=redux_path)

    # Calculate the S/N and coordinates
    rdxqa = ReductionAssessment('SNRG', drpf, clobber=clobber, analysis_path=analysis_path)

    # Peform the Voronoi binning to S/N>~10
    binned_spectra = SpatiallyBinnedSpectra('VOR10', drpf, rdxqa, clobber=clobber,
                                            analysis_path=analysis_path)

    # Fit the stellar kinematics
    stellar_continuum = StellarContinuumModel('GAU-MILESHC', binned_spectra, clobber=clobber,
                                              guess_vel=vel, guess_sig=100.,
                                              analysis_path=analysis_path)

    # Get the emission-line moments
    emission_line_moments = EmissionLineMoments('EMOMF', binned_spectra, clobber=clobber,
                                                stellar_continuum=stellar_continuum,
                                                redshift=nsa_redshift,
                                                analysis_path=analysis_path)

    # Get an estimate of the redshift of each bin using the first moment
    # of the H-alpha emission line:
    el_init_redshift = numpy.full(binned_spectra.nbins, nsa_redshift, dtype=float)
    # HARDCODED FOR A SPECIFIC EMISSION-LINE MOMENT DATABASE
    # TODO: Should
    #   - pass the EmissionLineMoments object to EmissionLineModel
    #   - include the channel used for this in EmissionLineModelDef
    halpha_channel = 7
    halpha_mom1_masked = emission_line_moments['ELMMNTS'].data['MASK'][:,halpha_channel] > 0
    # - Use the 1st moment of the H-alpha line
    el_init_redshift[ emission_line_moments['ELMMNTS'].data['BINID_INDEX'] ] \
                = emission_line_moments['ELMMNTS'].data['MOM1'][:,halpha_channel] \
                                / astropy.constants.c.to('km/s').value

    # - For missing bins in the moment measurements and bad H-alpha
    #   moment measurements, use the value for the nearest good bin
    bad_bins = numpy.append(emission_line_moments.missing_bins,
                    emission_line_moments['ELMMNTS'].data['BINID'][halpha_mom1_masked]).astype(int)
    if len(bad_bins) > 0:
        nearest_good_bin_index = binned_spectra.find_nearest_bin(bad_bins, indices=True)
        bad_bin_index = binned_spectra.get_bin_indices(bad_bins)
        el_init_redshift[bad_bin_index] = el_init_redshift[nearest_good_bin_index]

    # Setup the new emission-line fitter
    fitter = XJMCEmissionLineFitter()

    # Setup the new fitting method
    fit_method = EmissionLineModelDef('XJMC',       # Key for the fitting method
                                      0.0,          # Minimum S/N of the binned spectra to include
                                      None,         # Keyword for an artifact mask
                                      None,         # Keyword for an emission-line database
                                      fitter.par,   # Object with fit parameters
                                      fitter,       # Fitting class instance
                                      fitter.fit)   # Fitting function

    # Fit the emission lines
    emission_line_model = EmissionLineModel('XJMC',
                                            binned_spectra,
                                            stellar_continuum=stellar_continuum,
                                            redshift=el_init_redshift, dispersion=100.0,
                                            method_list=fit_method, 
                                            analysis_path=analysis_path, clobber=True)

    # The rest of this is just a single execution of the remaining
    # analysis steps in
    # $MANGADAP_DIR/python/mangadap/survey/manga_dap.py , with some
    # simplifications
#    spectral_indices = SpectralIndices('INDXEN', binned_spectra, redshift=nsa_redshift,
#                                       stellar_continuum=stellar_continuum,
#                                       emission_line_model=emission_line_model,
#                                       analysis_path=analysis_path,
#                                       clobber=False)
    spectral_indices = None

    construct_maps_file(drpf, rdxqa=rdxqa, binned_spectra=binned_spectra,
                        stellar_continuum=stellar_continuum,
                        emission_line_moments=emission_line_moments,
                        emission_line_model=emission_line_model,
                        spectral_indices=spectral_indices, nsa_redshift=nsa_redshift,
                        analysis_path=analysis_path)

    construct_cube_file(drpf, binned_spectra=binned_spectra, stellar_continuum=stellar_continuum,
                        emission_line_model=emission_line_model,
                        analysis_path=analysis_path)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))




import os
import time

from IPython import embed

import numpy

from astropy.io import fits
from matplotlib import pyplot

from mangadap.config import defaults
from mangadap.config import manga
from mangadap.datacube import MaNGADataCube
from mangadap.util.fitsutil import DAPFitsUtil
from mangadap.util.resolution import SpectralResolution
from mangadap.util.pixelmask import SpectralPixelMask

from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionmomentsdb import EmissionMomentsDB
from mangadap.par.emissionlinedb import EmissionLineDB

from mangadap.proc.templatelibrary import TemplateLibrary
from mangadap.proc.spectralstack import SpectralStack
from mangadap.proc.emissionlinemoments import EmissionLineMoments
from mangadap.proc.sasuke import Sasuke
from mangadap.proc.ppxffit import PPXFFit
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel, StellarContinuumModelBitMask
from mangadap.proc.emissionlinemodel import EmissionLineModelBitMask
from mangadap.proc.spectralfitting import EmissionLineFit

#-----------------------------------------------------------------------------

def get_redshift(plt, ifu, drpall_file=None):
    """
    Get the redshift of a galaxy from the DRPall file.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        drapall_file (:obj:`str`, optional):
            DRPall file. If None, attempts to use the default path to
            the file using environmental variables.
    
    Returns:
        :obj:`float`: The redshift to the galaxy observed by the
        provided PLATEIFU.
    """
    if drpall_file is None:
        drpall_file = manga.drpall_file()
    if not drpall_file.exists():
        raise FileNotFoundError(f'Could not find DRPall file: {drpall_file}')
    hdu = fits.open(str(drpall_file))
    indx = hdu[1].data['PLATEIFU'] == '{0}-{1}'.format(plt, ifu)
    return hdu[1].data['NSA_Z'][indx][0]


def get_spectra(plt, ifu, x, y, directory_path=None):
    """
    Extract spectra from a MaNGA observation.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        x (:obj:`int`, `numpy.ndarray`_):
            The spaxel coordinate along the RA axis.
        y (:obj:`int`, `numpy.ndarray`_):
            The spaxel coordinate along the DEC axis.
        directory_path (:obj:`str`, optional):
            Directory with the DRP LOGCUBE file. If None, uses the
            default directory path based on the environmental
            variables.

    Returns:
        :obj:`tuple`: Returns 4 numpy vectors: The wavelength, flux,
        flux inverse variance, and spectral resolution extracted from
        the datacube.
    """
    cube = MaNGADataCube.from_plateifu(plt, ifu, directory_path=directory_path)
    flat_indx = cube.spatial_shape[1]*x+y
    # This function always returns as masked array
    flux = cube.copy_to_masked_array(attr='flux', flag=cube.do_not_fit_flags())
    ivar = cube.copy_to_masked_array(attr='ivar', flag=cube.do_not_fit_flags())
    sres = cube.copy_to_array(attr='sres')
    return cube.wave, flux[flat_indx,:], ivar[flat_indx,:], sres[flat_indx,:]

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.perf_counter()


    # Plate-IFU to use
    plt = 7815
    ifu = 3702
    # Spaxel coordinates
    x = numpy.array([21,20,19])
    y = numpy.array([21,21,21])
    binid=numpy.array([0,0,1])
    # Number of spectra
    nspec = len(x)

    # Show the ppxf plots
#    fit_plots = True
    fit_plots = False

    # Show summary plots
    usr_plots = True #False

    # Template keywords
    sc_tpl_key = 'MILESHC'
    el_tpl_key = 'MILESHC'
#    el_tpl_key = 'BC03'

    # Emission-line database keywords
    elmom_key = 'ELBMILES'
    elfit_key = 'ELPMILES'
#    elmom_key = 'ELBTT'
#    elfit_key = 'ELPT4'

    # Template pixel scale a factor of 4 smaller than galaxy data
    velscale_ratio = 4

    # DAP source directory
    dapsrc = defaults.dap_source_dir()

    # Get the redshift
    drpver = 'v3_1_1'
    directory_path = dapsrc / 'data' / 'remote'
    drpall_file = directory_path / f'drpall-{drpver}.fits'
    # Assume all spectra have the same redshift
    z = numpy.array([get_redshift(plt, ifu, drpall_file)]*nspec)
    print('Redshift: {0}'.format(z[0]))
    dispersion = numpy.full_like(z, 100.)

    # Read the spectra
    print('reading spectra')
    wave, flux, ivar, sres = get_spectra(plt, ifu, x, y, directory_path=directory_path)
    ferr = numpy.ma.power(ivar, -0.5)

    # Fitting functions expect data to be in 2D arrays (for now):
    if len(flux.shape) == 1:
        flux = flux.reshape(1,-1)
        ferr = ferr.reshape(1,-1)
        sres = sres.reshape(1,-1)
        flux_binned = flux.copy()
        ferr_binned = ferr.copy()
        sres_binned = sres.copy()
        x_binned = x.copy()
        y_binned = y.copy()
        z_binned = z.copy()
        dispersion_binned = dispersion.copy()
    else:
        # Stack the spectra
        wave_binned, flux_binned, fsdev_binned, npix_binned, ivar_binned, sres_binned, \
                covar_binned = SpectralStack().stack(wave, flux, binid=binid, ivar=ivar, sres=sres)
        ferr_binned = numpy.ma.power(ivar_binned, -0.5)
        x_binned = numpy.array([numpy.mean(x[binid == i]) for i in numpy.unique(binid)])
        y_binned = numpy.array([numpy.mean(y[binid == i]) for i in numpy.unique(binid)])
        z_binned = numpy.array([numpy.mean(z[binid == i]) for i in numpy.unique(binid)])
        dispersion_binned = numpy.array([numpy.mean(dispersion)])

    if usr_plots:
        for f in flux:
            pyplot.plot(wave, f)
        pyplot.plot(wave_binned, flux_binned[0])
        pyplot.show()

    #-------------------------------------------------------------------
    # Fit the stellar continuum

    # Mask the 5577 sky line and the emission lines
    sc_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'),
                                      emldb=EmissionLineDB.from_key('ELPSCMSK'))

    # Construct the template library
    sc_tpl = TemplateLibrary(sc_tpl_key, match_resolution=False, velscale_ratio=velscale_ratio,
                             spectral_step=1e-4, log=True, hardcopy=False)
    sc_tpl_sres = numpy.mean(sc_tpl['SPECRES'].data, axis=0).ravel()

    # Instantiate the fitting class
    ppxf = PPXFFit(StellarContinuumModelBitMask())

    # Perform the fit
    cont_wave, cont_flux, cont_mask, cont_par \
        = ppxf.fit(sc_tpl['WAVE'].data.copy(), sc_tpl['FLUX'].data.copy(), wave_binned, 
                   flux_binned, ferr_binned, z_binned, dispersion_binned,
                   iteration_mode='no_global_wrej', reject_boxcar=100,
                   ensemble=False, velscale_ratio=velscale_ratio, mask=sc_pixel_mask,
                   matched_resolution=False, tpl_sres=sc_tpl_sres, obj_sres=sres_binned,
                   degree=8, moments=2, plot=fit_plots)

    # Remask the continuum fit
    sc_continuum = StellarContinuumModel.reset_continuum_mask_window(
                        numpy.ma.MaskedArray(cont_flux, mask=cont_mask>0))

    # Show the fit and residual
    if usr_plots:
        pyplot.plot(wave, flux_binned[0,:], label='Data')
        pyplot.plot(wave, sc_continuum[0,:], label='Model')
        pyplot.plot(wave, flux_binned[0,:] - sc_continuum[0,:], label='Resid')
        pyplot.legend()
        pyplot.xlabel('Wavelength')
        pyplot.ylabel('Flux')
        pyplot.show()
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # Get the emission-line moments using the fitted stellar continuum

    # Read the database that define the emission lines and passbands
    momdb = EmissionMomentsDB.from_key(elmom_key)

    # Measure the moments
    elmom = EmissionLineMoments.measure_moments(momdb, wave, flux_binned, continuum=sc_continuum,
                                                redshift=z_binned)
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # Fit the emission-line model

    # Set the emission-line continuum templates if different from those
    # used for the stellar continuum
    if sc_tpl_key == el_tpl_key:
        # If the keywords are the same, just copy over the previous
        # library ...
        el_tpl = sc_tpl
        el_tpl_sres = sc_tpl_sres
        # ... and the best fitting stellar kinematics
        stellar_kinematics = cont_par['KIN']
    else:
        # If the template sets are different, we need to match the
        # spectral resolution to the galaxy data ...
        _sres = SpectralResolution(wave, sres[0,:], log10=True)
        el_tpl = TemplateLibrary(el_tpl_key, sres=_sres, velscale_ratio=velscale_ratio,
                                 spectral_step=1e-4, log=True, hardcopy=False)
        el_tpl_sres = numpy.mean(el_tpl['SPECRES'].data, axis=0).ravel()
        # ... and use the corrected velocity dispersions.
        stellar_kinematics = cont_par['KIN']
        stellar_kinematics[:,1] = numpy.ma.sqrt(numpy.square(cont_par['KIN'][:,1]) -
                                                    numpy.square(cont_par['SIGMACORR_EMP']))
    
    # Mask the 5577 sky line
    el_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'))

    # Read the emission line fitting database
    emldb = EmissionLineDB.from_key(elfit_key)

    # Instantiate the fitting class
    emlfit = Sasuke(EmissionLineModelBitMask())

    # Perform the emission-line fit on each spectrum using the stellar
    # kinematics from the stacked spectrum
    eml_wave, model_flux, eml_flux, eml_mask, eml_fit_par, eml_eml_par \
            = emlfit.fit(emldb, wave_binned, flux_binned, obj_ferr=ferr_binned,
                         obj_mask=el_pixel_mask, obj_sres=sres_binned,
                         guess_redshift=z_binned, guess_dispersion=dispersion_binned,
                         reject_boxcar=101, stpl_wave=el_tpl['WAVE'].data,
                         stpl_flux=el_tpl['FLUX'].data, stpl_sres=el_tpl_sres,
                         stellar_kinematics=stellar_kinematics, etpl_sinst_mode='offset',
                         etpl_sinst_min=10., velscale_ratio=velscale_ratio,
                         matched_resolution=False, mdegree=8, plot=fit_plots,
                         remapid=binid, remap_flux=flux, remap_ferr=ferr,
                         remap_mask=el_pixel_mask, remap_sres=sres, remap_skyx=x, remap_skyy=y,
                         obj_skyx=x_binned, obj_skyy=y_binned)

    # Line-fit metrics
    eml_eml_par = EmissionLineFit.line_metrics(emldb, wave, flux, ferr, model_flux, eml_eml_par,
                                               model_mask=eml_mask, bitmask=emlfit.bitmask)

    # Get the stellar continuum that was fit for the emission lines
    elcmask = eml_mask > 0
    edges = numpy.ma.notmasked_edges(numpy.ma.MaskedArray(model_flux, mask=elcmask), axis=1)
    for i,s,e in zip(edges[0][0],edges[0][1],edges[1][1]):
        elcmask[i,s:e+1] = False
    el_continuum = numpy.ma.MaskedArray(model_flux - eml_flux, mask=elcmask)

    # Plot the result
    if usr_plots:
        for i in range(flux.shape[0]):
            pyplot.plot(wave, flux[i,:], label='Data')
            pyplot.plot(wave, model_flux[i,:], label='Model')
            pyplot.plot(wave, el_continuum[i,:], label='EL Cont.')
            pyplot.plot(wave, sc_continuum[binid[i],:], label='SC Cont.')
            pyplot.legend()
            pyplot.xlabel('Wavelength')
            pyplot.ylabel('Flux')
            pyplot.show()

    # Remeasure the emission-line moments with the new continuum
    new_elmom = EmissionLineMoments.measure_moments(momdb, wave, flux, continuum=el_continuum,
                                                    redshift=z)

    # Compare the summed flux and Gaussian-fitted flux for all the
    # fitted lines
    if usr_plots:
        pyplot.scatter(emldb['restwave'], new_elmom['FLUX'][0,:]-eml_eml_par['FLUX'][0,:],
                       c=eml_eml_par['FLUX'][0,:], cmap='viridis', marker='.', s=60, lw=0,
                       zorder=4)
        pyplot.grid()
        pyplot.xlabel('Wavelength')
        pyplot.ylabel('Summed-Gaussian Difference')
        pyplot.show()

    # TODO: Add the spectral index calls...

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




from pathlib import Path
import os
import time
import argparse

from IPython import embed

import numpy

from astropy.io import fits

from matplotlib import pyplot
from matplotlib.ticker import NullFormatter
from matplotlib.widgets import Button, AxesWidget, Cursor, RangeSlider

from mangadap.dapfits import DAPCubeBitMask
from mangadap.util.bitmask import BitMask
from mangadap.config import defaults

from mangadap.scripts import scriptbase


#-----------------------------------------------------------------------
# Pointer class
class Pointer(AxesWidget):
    """
    """
    def __init__(self, ax, pos, **kwargs):
        AxesWidget.__init__(self, ax)

        self.cursor = Cursor(ax, useblit=True, **kwargs)

        self.connect_event('button_press_event', self._update)
        self.connect_event('button_release_event', self._update)
        self.cnt = 0
        self.observers = {}
        self.drag_active = False
        self.pos = pos

    def _update(self, event):
        """update the pointer position"""
        if self.ignore(event):
            return

        if event.button != 1:
            return

        if event.name == 'button_press_event' and event.inaxes is self.ax:
            self.drag_active = True
            event.canvas.grab_mouse(self.ax)

        if not self.drag_active:
            return

        elif event.name == 'button_release_event' \
                or (event.name == 'button_press_event' and event.inaxes is not self.ax):
            self.drag_active = False
            event.canvas.release_mouse(self.ax)
            return

        self.set_val((event.xdata, event.ydata))

    def set_val(self, val):
        self.pos = val
        if not self.eventson:
            return
        for cid, func in self.observers.items():
            func(self.pos)

    def on_changed(self, func):
        """
        When the pointer position is changed, call *func* with the new
        position.

        A connection id is returned which can be used to disconnect
        """
        cid = self.cnt
        self.observers[cid] = func
        self.cnt += 1
        return cid

    def disconnect(self, cid):
        """remove the observer with connection id *cid*"""
        try:
            del self.observers[cid]
        except KeyError:
            pass

#-----------------------------------------------------------------------------

class MaNGATargetBitMask(BitMask):
    r"""
    Structure with the DRP mask bits.
    """
    def __init__(self, sdss_maskbits=None, trgt='MANGA_TARGET1'):
        _sdss_maskbits = defaults.sdss_maskbits_file() if sdss_maskbits is None else sdss_maskbits
        tmp = BitMask.from_par_file(str(_sdss_maskbits), trgt)
        keys, descr = tmp._init_objs()
        super(MaNGATargetBitMask, self).__init__(keys, descr=descr)


def channel_bitmask():
    return BitMask(['ERRORS', 'MASKED'])


def pull_maps(hdu, ext, toerr=False):

    # Copy the map data
    maps = hdu[ext].data
    if maps.ndim not in [ 2, 3 ]:
        raise ValueError('WTF')

    bm = channel_bitmask()
    bits = numpy.zeros(maps.shape[0] if maps.ndim == 3 else 1, dtype=bm.minimum_dtype())

    try:
        err_ext = hdu[ext].header['ERRDATA']
        maps_err = numpy.ma.power(hdu[err_ext].data, -0.5) if toerr else hdu[err_ext].data
        bits = bm.turn_on(bits, 'ERRORS')
    except:
        maps_err = numpy.zeros(maps.shape, dtype=float)

    try:
        msk_ext = hdu[ext].header['QUALDATA']
        maps_msk = hdu[msk_ext].data
        bits = bm.turn_on(bits, 'MASKED')
    except:
        maps_msk = numpy.zeros(maps.shape, dtype=numpy.uint8)

    if len(maps.shape) == 2:
        return numpy.array([ ext ]), bits, maps.reshape(1,*maps.shape), \
                    maps_err.reshape(1,*maps.shape), maps_msk.reshape(1,*maps.shape)

    nmaps = maps.shape[0]
    names = numpy.full(nmaps, ext, dtype=str)
    ndig = int(numpy.log10(nmaps))+1

    names = numpy.array([ '{0}:{1}'.format(ext, hdu[ext].header['C'+'{0}'.format(i+1).zfill(ndig)])
                            for i in range(nmaps) ])

    return names, bits, maps, maps_err, maps_msk


def build_maps(hdu, ext=['BIN_MFLUX', 'BIN_SNR', 'STELLAR_VEL', 'STELLAR_SIGMA']):

    # TODO: Deal with the transpose

    # Always get the bin ID extension
    binid = hdu['BINID'].data

    # Build the maps cube
    _ext = numpy.array([ 'BIN_MFLUX', 'BIN_SNR', 'STELLAR_VEL', 'STELLAR_SIGMA']) \
                if ext is None else numpy.asarray(ext)

    print('Building maps from extensions: {0}'.format(_ext))

    ndim = numpy.array([ hdu[e].data.ndim for e in _ext ])
    spatial_shape = hdu[_ext[ndim==2][0]].shape if numpy.sum(ndim == 2) > 0 \
                            else hdu[_ext[ndim==3][0]].shape[1:3]

    channel_name = numpy.empty(0, dtype=object)
    channel_bits = numpy.empty(0, dtype=numpy.uint8)
    map_cube = numpy.empty((0,*spatial_shape), dtype=float)
    map_cube_ivar = numpy.empty((0,*spatial_shape), dtype=float)
    map_cube_mask = numpy.empty((0,*spatial_shape), dtype=numpy.uint64)
    names = numpy.empty(0, dtype=object)
    zmin = numpy.empty(0, dtype=float)
    zmax = numpy.empty(0, dtype=float)
    for i in range(len(_ext)):
        _names, _bits, _maps, _maps_ivar, _maps_mask = pull_maps(hdu, _ext[i])
        channel_name = numpy.append( channel_name, _names )
        channel_bits = numpy.append( channel_bits, _bits )
        map_cube = numpy.append(map_cube, _maps, axis=0)
        map_cube_ivar = numpy.append(map_cube_ivar, _maps_ivar, axis=0)
        map_cube_mask = numpy.append(map_cube_mask, _maps_mask, axis=0)
        _maps = numpy.ma.MaskedArray(_maps, mask=_maps_mask>0) \
                    if numpy.sum(_maps_mask>0) < _maps_mask.size else numpy.ma.MaskedArray(_maps)
        mins = numpy.ma.amin(_maps) if _maps.ndim == 2 else numpy.ma.amin(_maps, axis=(1,2))
        maxs = numpy.ma.amax(_maps) if _maps.ndim == 2 else numpy.ma.amax(_maps, axis=(1,2))
        zmin = numpy.append(zmin, mins)
        zmax = numpy.append(zmax, maxs)

    return binid, channel_name, channel_bits, map_cube, map_cube_ivar, map_cube_mask, zmin, zmax


def build_model_spectra(hdu, masked=True):

    if masked:
        bm = DAPCubeBitMask()
        masked_spec = numpy.ma.MaskedArray(hdu['FLUX'].data,
                                           mask=bm.flagged(hdu['MASK'].data,
                                                           flag=['IGNORED', 'FORESTAR',
                                                                 'FLUXINVALID', 'IVARINVALID']))

        model_spec = numpy.ma.MaskedArray(hdu['MODEL'].data,
                                          mask=bm.flagged(hdu['MASK'].data,
                                                          flag=['FITIGNORED', 'FITFAILED']))

        model_emission = numpy.ma.MaskedArray(hdu['MODEL'].data - hdu['STELLAR'].data,
                                              mask=bm.flagged(hdu['MASK'].data,
                                                              flag=['FITIGNORED', 'FITFAILED',
                                                                    'ELIGNORED', 'ELFAILED']))

        model_continuum = numpy.ma.MaskedArray(hdu['MODEL'].data - hdu['EMLINE'].data,
                                               mask=bm.flagged(hdu['MASK'].data,
                                                               flag=['FITIGNORED', 'FITFAILED']))

    else:
        masked_spec = hdu['FLUX'].data
        model_spec = hdu['MODEL'].data
        model_emission = hdu['MODEL'].data - hdu['STELLAR'].data
        model_continuum = hdu['MODEL'].data - hdu['EMLINE'].data

    return hdu['WAVE'].data, masked_spec, model_spec, model_continuum, model_emission

                                      
def select_spectrum(spec, pos):
    return spec[:,pos[1],pos[0]]


class ImageMarker:
    def __init__(self, s, pointer, pos=None):
        self.s = s
        self.pointer = pointer
        self.reset_pos(pos)
    def __call__(self, val=None):
        self.apply_offsets()
        pyplot.draw()
    def reset_pos(self, pos=None):
        self.pos = tuple([int(numpy.round(p)) for p in self.pointer.pos]) if pos is None else pos
    def apply_offsets(self):
        self.reset_pos()
        self.s.set_offsets(self.pos)


class UpdateableRangeSlider(RangeSlider):
    def __init__(self, ax, label, valmin, valmax, valinit=None, valfmt=None, closedmin=True,
                 closedmax=True, dragging=True, valstep=None, orientation="horizontal",
                 track_color='lightgrey', handle_style=None, **kwargs):

        super().__init__(ax, label, valmin, valmax, valinit=valinit, valfmt=valfmt,
                         closedmin=closedmin, closedmax=closedmax, dragging=dragging,
                         valstep=valstep, orientation=orientation, track_color=track_color,
                         handle_style=handle_style, **kwargs)
        # Save the axes reference internally
        self.ax = ax
        self.label.set_position((0.5, 1.0))
        self.label.set_verticalalignment('bottom')
        self.label.set_horizontalalignment('center')
        self.label.set_weight('bold')

    def update_range(self, rng, label):
        self._active_handle = None
        xy = self.poly.get_xy()
        xy[:,0] = numpy.roll(numpy.repeat(rng, 3)[:-1],-1)
        self.poly.set_xy(xy)
        self.valmin, self.valmax = rng
        self.valinit = numpy.array(rng)
        self._handles[0].set_xdata(numpy.array([rng[0]]))
        self._handles[1].set_xdata(numpy.array([rng[1]]))
        self.ax.set_xlim(rng)
        self.label.set_text(label)
        self.set_val(rng)


class UpdateableMap:
    def __init__(self, fig, im, cbar, channel_names, channel_bits, map_cube, map_cube_ivar,
                 map_cube_mask, zmin, zmax, current=0):
        self.fig = fig
        self.im = im
        self.cbar = cbar
        self.bitmask = channel_bitmask()

        self.current = current
        self.nmap = map_cube.shape[0]
        self.channel_names = channel_names
        self.channel_bits = channel_bits
        self.map_cube = map_cube
        self.map_cube_ivar = map_cube_ivar
        self.map_cube_mask = map_cube_mask
        self.zmin = zmin
        self.zmax = zmax

        self.msk_zmin = zmin.copy()
        self.msk_zmax = zmax.copy()
        indx = self.bitmask.flagged(self.channel_bits, flag='MASKED')
        self.map_data = numpy.ma.MaskedArray(self.map_cube,
                                             mask=map_cube_mask>0).reshape(-1,len(indx))[:,indx]

        self.snr_zmin = zmin.copy()
        self.snr_zmax = zmax.copy()
        indx = self.bitmask.flagged(self.channel_bits, flag='ERRORS')
        self.map_data = numpy.ma.MaskedArray(self.map_cube*numpy.sqrt(self.map_cube_ivar),
                                             mask=map_cube_mask>0).reshape(-1,len(indx))[:,indx]
        self.snr_zmin[indx] = numpy.ma.amin(self.map_data, axis=0)
        self.snr_zmax[indx] = numpy.ma.amax(self.map_data, axis=0)

        self.map_data = map_cube.copy()
        self.rng = (self.msk_zmin[self.current], self.msk_zmax[self.current])

        # Colorbar range slider
        self.slider_ax = pyplot.axes([0.02, 0.08, 0.7/2.5, 0.04])
        self.zslider = UpdateableRangeSlider(self.slider_ax, self.channel_names[self.current],
                                             self.rng[0], self.rng[1], valinit=self.rng)
        self.zslider.on_changed(self.change_range)

        self.masked = False
        self.snr = False

        mask_button_ax = fig.add_axes([0.015+0*0.13/2.5, 0.02, 0.13/2.5, 0.05])
        self.mask_button = Button(mask_button_ax, 'Mask', color='0.7', hovercolor='0.9')
        self.mask_button.on_clicked(self.toggle_mask)

        snr_button_ax = fig.add_axes([0.015+1*0.13/2.5, 0.02, 0.13/2.5, 0.05])
        self.snr_button = Button(snr_button_ax, 'SNR', color='0.7', hovercolor='0.9')
        self.snr_button.on_clicked(self.toggle_snr)

        frst_button_ax = fig.add_axes([0.015+2*0.13/2.5+0.01, 0.02, 0.13/2.5, 0.05])
        self.frst_button = Button(frst_button_ax, 'First', color='0.7', hovercolor='0.8')
        self.frst_button.on_clicked(self.first)

        prev_button_ax = fig.add_axes([0.015+3*0.13/2.5+0.01, 0.02, 0.13/2.5, 0.05])
        self.prev_button = Button(prev_button_ax, 'Prev', color='0.7', hovercolor='0.8')
        self.prev_button.on_clicked(self.prev)

        next_button_ax = fig.add_axes([0.015+4*0.13/2.5+0.01, 0.02, 0.13/2.5, 0.05])
        self.next_button = Button(next_button_ax, 'Next', color='0.7', hovercolor='0.8')
        self.next_button.on_clicked(self.next)

        last_button_ax = fig.add_axes([0.015+5*0.13/2.5+0.01, 0.02, 0.13/2.5, 0.05])
        self.last_button = Button(last_button_ax, 'Last', color='0.7', hovercolor='0.8')
        self.last_button.on_clicked(self.last)

        pyplot.draw()


    def get_masked_data(self):
        return numpy.ma.MaskedArray(self.map_cube.copy(), mask=self.map_cube_mask>0)
    def get_snr_data(self):
        return self.map_cube*numpy.sqrt(self.map_cube_ivar)
    def change_range(self, val):
        self.im.set_clim(*val)
        pyplot.draw()
    def toggle_mask(self, val=None):
        if self.masked:
            self.masked = False
            self.map_data = self.get_snr_data() if self.snr else self.map_cube.copy()
        else:
            self.masked = True
            self.map_data = self.get_snr_data() if self.snr else self.map_cube.copy()
            self.map_data = numpy.ma.MaskedArray(self.map_data, mask=self.map_cube_mask>0)
        t = self.mask_button.hovercolor
        self.mask_button.hovercolor = self.mask_button.color
        self.mask_button.color = t
        self.update_map()
    def toggle_snr(self, val=None):
        if self.snr:
            self.snr = False
            self.map_data = self.get_masked_data() if self.masked else self.map_cube.copy()
        else:
            self.snr = True
            self.map_data = self.get_masked_data() if self.masked else self.map_cube.copy()
            self.map_data = self.map_data*numpy.sqrt(self.map_cube_ivar)
        t = self.snr_button.hovercolor
        self.snr_button.hovercolor = self.snr_button.color
        self.snr_button.color = t
        self.update_map()
    def first(self, val=None):
        self.current = 0
        self.update_map()
    def prev(self, val=None):
        self.current = self.nmap-1 if self.current == 0 else self.current - 1
        self.update_map()
    def next(self, val=None):
        self.current = 0 if self.current == self.nmap-1 else self.current + 1
        self.update_map()
    def last(self, val=None):
        self.current = self.nmap-1
        self.update_map()

    def update_map(self):
        self.im.set_data(self.map_data[self.current])
        if self.snr:
            self.rng = (self.snr_zmin[self.current], self.snr_zmax[self.current])
        else:
            self.rng = (self.msk_zmin[self.current], self.msk_zmax[self.current])
        self.im.set_clim(*self.rng)
        self.zslider.update_range(self.rng, self.channel_names[self.current])
        pyplot.draw()


class UpdateSpectrumPlot:
    def __init__(self, binid, masked_spec, model_continuum, model_emission, model_spec, specl,
                 contl, fitl, continuum_residl, em_fitl, residl, image_marker, spx_coo_txt,
                 binid_txt):

        self.binid = binid
        self.masked_spec = masked_spec
        self.model_continuum = model_continuum
        self.model_emission = model_emission
        self.model_spec = model_spec
        self.specl = specl
        self.contl = contl
        self.fitl = fitl
        self.cresidl = continuum_residl
        self.em_fitl = em_fitl
        self.residl = residl
        self.marker = image_marker
        self.cootxt = spx_coo_txt
        self.bintxt = binid_txt

    def __call__(self, val=None):
        self.marker.apply_offsets()
        _masked_spec = select_spectrum(self.masked_spec, self.marker.pos)
        _model_spec = select_spectrum(self.model_spec, self.marker.pos)
        _model_continuum = select_spectrum(self.model_continuum, self.marker.pos)
        _model_emission = select_spectrum(self.model_emission, self.marker.pos)
        self.specl.set_ydata(_masked_spec)
        self.contl.set_ydata(_model_continuum)
        self.fitl.set_ydata(_model_spec)
        self.cresidl.set_ydata(_masked_spec - _model_continuum)
        self.em_fitl.set_ydata(_model_emission)
        self.residl.set_ydata(_masked_spec - _model_spec)

        self.cootxt.set_text(f'Spaxel = {self.marker.pos}')
        self.bintxt.set_text(f'Bin IDs = {self.binid[:,self.marker.pos[1], self.marker.pos[0]]}')

        pyplot.draw()
        

#def parse_plt_ifu(maps_file):
#    file_name = maps_file.split('/')[-1]
#    return tuple([ int(n) for n in file_name.split('-')[1:3] ])

def parse_name_method(maps_file):
    f = Path(maps_file).resolve()
    if not f.exists():
        raise FileNotFoundError(f'{f} does not exist!')
    components = f.name.split('-')
    components[-1] = components[-1].split('.')[0]
    i = components.index('MAPS')
    return '-'.join(components[:i]), '-'.join(components[i+1:])


def manga_dap_inspector(maps_file, model_file, ext=None, masked_spectra=True):

    # Try to get the file identifier and DAP method
    name, method = parse_name_method(maps_file)

    #-------------------------------------------------------------------
    # Read the maps to plot
    print('Reading MAP file: {0}'.format(maps_file))
    with fits.open(maps_file) as hdu:
        # Get the map data
        binid, channel_names, channel_bits, map_cube, map_cube_ivar, map_cube_mask, \
                map_zmin, map_zmax = build_maps(hdu, ext=ext)

        # Get the MaNGA ID
        mangaid = hdu['PRIMARY'].header['MANGAID'] if 'MANGAID' in hdu['PRIMARY'].header else None

        # Determine the sample flags
        if 'MNGTARG1' in hdu['PRIMARY'].header:
            trg_flags = MaNGATargetBitMask().flagged_bits(hdu['PRIMARY'].header['MNGTARG1'])
        else:
            trg_flags = None

    target_type = 'None'
    if trg_flags is not None:
        if 'SECONDARY_v1_2_0' in trg_flags:
            target_type = 'S'
        if 'PRIMARY_v1_2_0' in trg_flags:
            target_type = 'P'
        if 'COLOR_ENHANCED_v1_2_0' in trg_flags:
            target_type = 'P+'

    #-------------------------------------------------------------------
    # Read the spectra and models to plot
    print('Reading MODEL file: {0}'.format(model_file))
    with fits.open(model_file) as hdu:
        wave, masked_spec, model_spec, model_continuum, model_emission \
                = build_model_spectra(hdu, masked=masked_spectra)

    # Start the plot
    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(2.5*w,h))

    # ------------------------------------------------------------------
    # Create the channel image plot
    on_sky_xlim = numpy.array([0, map_cube.shape[2]]) - 0.5
    on_sky_ylim = numpy.array([0, map_cube.shape[1]]) - 0.5
    image_ax = fig.add_axes([0.02, 0.25, 0.7/2.5, 0.7], facecolor='0.95')
    image_ax.tick_params(axis='both', which='both', direction='in')
    image_cax = fig.add_axes([0.02+0.7/2.5+0.01, 0.25, 0.01, 0.7])
    image_ax.set_xlim(on_sky_xlim)
    image_ax.set_ylim(on_sky_ylim)
    image_ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')

    # Create the initial plot, color bar, and labels
    im = image_ax.imshow(map_cube[0], origin='lower', interpolation='nearest',
                         cmap='RdBu_r', vmin=map_zmin[0], vmax=map_zmax[0], zorder=3)
    cbar = pyplot.colorbar(im, cax=image_cax)

    if mangaid is None:
        image_ax.text(0.5, 1.03, f'{name} {method}', ha='center', va='center',
                      transform=image_ax.transAxes)
    else:
        image_ax.text(0.5, 1.03, f'{name}; MaNGA ID={mangaid}; {target_type}', ha='center',
                      va='center', transform=image_ax.transAxes)
    image_ax.set_ylabel('Y [pix]')
    image_ax.set_xlabel('X [pix]')

    # Start the marker at the center of the map
    pos = tuple(*(numpy.array([map_cube.shape[1:][::-1]])/2).astype(int))
    marker = image_ax.scatter(*pos, marker='s', s=20, color='k', alpha=0.5, zorder=4)
    # ------------------------------------------------------------------


    # ------------------------------------------------------------------
    # Add the buttons and sliders for changing the map properties
    map_selector = UpdateableMap(fig, im, cbar, channel_names, channel_bits, map_cube,
                                 map_cube_ivar, map_cube_mask, map_zmin, map_zmax)
    # ------------------------------------------------------------------


    # ------------------------------------------------------------------
    # Add the image marker
    coo_pointer = Pointer(im.axes, pos, color='C3', lw=0.5, alpha=0.5)
    coo_selector = ImageMarker(marker, coo_pointer)
    # ------------------------------------------------------------------

    # ------------------------------------------------------------------
    # Add the spectrum plots
    _masked_spec = select_spectrum(masked_spec, pos)
    _model_spec = select_spectrum(model_spec, pos)
    _model_continuum = select_spectrum(model_continuum, pos)
    _model_emission = select_spectrum(model_emission, pos)

    wavelim = [wave[0], wave[-1]]
    fluxlim = [numpy.ma.amin(masked_spec), numpy.ma.amax(masked_spec)]
    continuum_residlim = [numpy.ma.amin(masked_spec-model_continuum)-0.1,
                          numpy.ma.amax(masked_spec-model_continuum)+0.1]
    residlim = [numpy.ma.amin(masked_spec-model_spec)-0.1,
                numpy.ma.amax(masked_spec-model_spec)+0.1]

#    fluxlim = [numpy.ma.amin(_masked_spec)-1, numpy.ma.amax(_masked_spec)+1]
#    continuum_residlim = [numpy.ma.amin(_masked_spec-_model_continuum)-0.1,
#                          numpy.ma.amax(_masked_spec-_model_continuum)+0.1]
#    residlim = [numpy.ma.amin(_masked_spec-_model_spec)-0.1,
#                numpy.ma.amax(_masked_spec-_model_spec)+0.1]

    spectrum_ax = fig.add_axes([ 0.40, 0.31, 0.58, 0.6 ], facecolor='0.95')
    spectrum_ax.set_xlim(wavelim)
    spectrum_ax.set_ylim(fluxlim)
    spectrum_ax.minorticks_on()
    spectrum_ax.tick_params(axis='both', which='both', direction='in')
    spectrum_ax.tick_params(which='major', length=6)
    spectrum_ax.tick_params(which='minor', length=3)
    spectrum_ax.xaxis.set_major_formatter(NullFormatter())
    spectrum_ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
    spectrum_ax.grid(True, which='minor', color='0.85', zorder=0, linestyle=':')

    wave_ax = spectrum_ax.twiny()
    wave_ax.set_xlim(wavelim)
    wave_ax.minorticks_on()
    wave_ax.tick_params(axis='both', which='both', direction='in')
    wave_ax.tick_params(which='major', length=6)
    wave_ax.tick_params(which='minor', length=3)

    continuum_resid_ax = fig.add_axes([ 0.40, 0.11, 0.58, 0.20 ], facecolor='0.95',
                                        sharex=spectrum_ax)
    continuum_resid_ax.set_xlim(wavelim)
    continuum_resid_ax.set_ylim(continuum_residlim)
    continuum_resid_ax.minorticks_on()
    continuum_resid_ax.tick_params(axis='both', which='both', direction='in')
    continuum_resid_ax.tick_params(which='major', length=6)
    continuum_resid_ax.tick_params(which='minor', length=3)
    continuum_resid_ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
    continuum_resid_ax.grid(True, which='minor', color='0.85', zorder=0, linestyle=':')

    
    resid_ax = fig.add_axes([ 0.40, 0.01, 0.58, 0.10 ], facecolor='0.95', sharex=spectrum_ax)
    resid_ax.set_xlim(wavelim)
    resid_ax.set_ylim(residlim)
    resid_ax.minorticks_on()
    resid_ax.tick_params(axis='both', which='both', direction='in')
    resid_ax.tick_params(which='major', length=6)
    resid_ax.tick_params(which='minor', length=3)
    resid_ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
    resid_ax.grid(True, which='minor', color='0.85', zorder=0, linestyle=':')

    
    spec, = spectrum_ax.step(wave, _masked_spec, where='mid', lw=0.5, color='k', zorder=3)
    cont, = spectrum_ax.plot(wave, _model_continuum, lw=0.5, color='b', zorder=4)
    fit, = spectrum_ax.plot(wave, _model_spec, lw=0.5, color='g', zorder=5)

    continuum_resid, = continuum_resid_ax.step(wave, _masked_spec - _model_continuum,
                                               where='mid', lw=0.5, color='k', zorder=3)
    em_fit, = continuum_resid_ax.plot(wave, _model_emission, lw=0.5, color='r', zorder=4)

    resid, = resid_ax.step(wave, _masked_spec - _model_spec,
                           where='mid', lw=0.5, color='k',zorder=3)

    spectrum_ax.text(0.5, 1.1, r'Wavelength ($\AA$)', horizontalalignment='center',
                     verticalalignment='center', transform=spectrum_ax.transAxes)
    spectrum_ax.text(1.02, 0.5, 'Flux', horizontalalignment='center', verticalalignment='center',
                     transform=spectrum_ax.transAxes, rotation='vertical')
    continuum_resid_ax.text(1.02, 0.5, r'Cont. $\Delta$', horizontalalignment='center',
                            verticalalignment='center', transform=continuum_resid_ax.transAxes,
                            rotation='vertical')
    resid_ax.text(1.02, 0.5, r'$\Delta$', horizontalalignment='center',
                  verticalalignment='center', transform=resid_ax.transAxes, rotation='vertical')

    spx_coo_txt = spectrum_ax.text(0.98, 0.92, 'Spaxel = {0}'.format(coo_selector.pos),
                                   horizontalalignment='right', verticalalignment='center',
                                   transform=spectrum_ax.transAxes, fontsize=8)
    binid_txt = spectrum_ax.text(0.98, 0.87, 'Bin IDs = {0}'.format(
                                        binid[:,coo_selector.pos[1],coo_selector.pos[0]]),
                                 horizontalalignment='right', verticalalignment='center',
                                 transform=spectrum_ax.transAxes, fontsize=8)

    spectrum_selector = UpdateSpectrumPlot(binid, masked_spec, model_continuum, model_emission,
                                           model_spec, spec, cont, fit, continuum_resid, em_fit,
                                           resid, coo_selector, spx_coo_txt, binid_txt)
    coo_pointer.on_changed(spectrum_selector)
    # ------------------------------------------------------------------

    pyplot.show()


class MangaDapInspector(scriptbase.ScriptBase):

    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'manga_dap_inspector'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Inspect results of the MaNGA DAP', width=width)

        parser.add_argument('--maps_file', type=str, default=None, help='Output MAPS file')
        parser.add_argument('--model_file', type=str, default=None, help='Output LOGCUBE model file')

        parser.add_argument('--ext', type=str, nargs='*', help='List of map extensions to include',
                            default=None)

        parser.add_argument('--maskspec', action='store_true', default=False,
                            help='Use masks for spectra.')
        return parser

    @staticmethod
    def main(args):

        # Can it proceed?
        maps_file, model_file = args.maps_file, args.model_file
        if maps_file is None or model_file is None:
            raise ValueError('Must provide both a MAPS file and a model LOGCUBE file.')

        manga_dap_inspector(maps_file, model_file, ext=args.ext, masked_spectra=args.maskspec)





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""
Provides a set of functions that define and return defaults used by the
MaNGA DAP, such as paths and file names.

----        

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

from pathlib import Path
from pkg_resources import resource_filename


def dap_source_dir():
    from pkg_resources import resource_filename
    return Path(resource_filename('mangadap', '')).resolve()


def dap_data_root():
    """Return the root directory with the DAP data."""
    return Path(resource_filename('mangadap', 'data')).resolve()


def dap_config_root():
    """Return the root directory with the DAP config data."""
    return Path(resource_filename('mangadap', 'config')).resolve()


def sdss_maskbits_file():
    """Return the path to the sdss maskbits yanny file."""
    maskbits_file = dap_data_root() / 'sdss' / 'sdssMaskbits.par'
    return maskbits_file if maskbits_file.exists() else None





"""
Defines the class used to automate the batch execution of the MaNGA DAP.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
from pathlib import Path
import time
import shutil
import warnings

from IPython import embed

import numpy

try:
    import pbs.queue
except:
    warnings.warn('Could not import pbs.queue!  Any cluster submission will fail!', ImportWarning)

from ..config import manga, manga_environ, python_versions
from .drpcomplete import DRPComplete
from ..util.parser import arginp_to_list
from ..util.fileio import create_symlink


class rundap:
    r"""
    Class used for automated execution of the MaNGA DAP via submission
    of scripts to a cluster.

    [ More complete explanation of the class here. ]

    .. note::
    
        This is purely a survey-level utility for running the DAP at
        Utah, including submitting jobs to the cluster.  No user-level
        usage of the DAP should be considered here, apart from not
        hindering such usage of the primary python programs!

    .. todo::

        - Have :func:`_read_arg` return the boolean for :attr:`version`
          instead of keeping it as an attribute.
        - Ditch ``console`` and :func:`_read_arg` in favor of the
          more typical ``parser`` and ``main`` pairing.

    Args:
        overwrite (:obj:`bool`, optional):
            Overwrite and redo any existing analysis.
        console (:obj:`bool`, optional):
            The class has been executed from the terminal with
            command-line arguments that should be parsed.
        quiet (:obj:`bool`, optional):
            Suppress output
        print_version (:obj:`bool`, optional):
            Print the class version and return.
        drpver (:obj:`str`, optional):
            DRP version to analyze.  Default is defined by
            :func:`mangadap.config.manga.drp_version`
        redux_path (:obj:`str`, optional):
            The top-level path with the DRP files used to override the
            default defined by
            :func:`mangadap.config.manga.drp_redux_path`.
        dapver (:obj:`str`, optional):
            The DAP version to use for the analysis, used to override
            the default defined by
            :func:`mangadap.config.manga.dap_version`.  *Only
            used to define the file paths, not used to switch to the
            specified code version!*
        analysis_path (:obj:`str`, optional):
            The top-level path for the DAP output files, used to
            override the default defined by
            :func:`mangadap.config.manga.dap_analysis_path`.
        plan_file (:obj:`str`, optional):
            Name of the plan file to use for *all* DRP data in this run,
            used to override the default defined by
            :func:`mangadap.par.analysisplan.AnalysisPlanSet.default`.
        platelist (:obj:`int`, :obj:`str`, optional):
            List of plates to analyze; default is to search the
            :attr:`redux_path` for any DRP files; see
            :class:`mangadap.survey.drpcomplete.DRPComplete`.
        ifudesignlist (:obj:`int`, :obj:`str`, optional):
            List of ifudesigns to analyze: default is to search the
            :attr:`redux_path` for any DRP files; see
            :class:`mangadap.survey.drpcomplete.DRPComplete`.
        combinatorics (:obj:`bool`, optional): 
            Use all unique combinations of the entered
            plate/ifudesign/mode lists to create the full list of DRP
            files to analyze.
        list_file (:obj:`str`, optional):
            File with a list of plates and ifudesigns to analyze. The
            file must have two columns, with one plate-ifu
            combination per file row.
        sres_ext (:obj:`str`, optional):
            The extension to use when constructing the
            spectral resolution vectors for the MaNGA datacubes. See
            :func:`mangadap.datacube.manga.MaNGADataCube.spectral_resolution`.
        sres_fill (:obj:`bool`, optional):
            Fill masked spectral-resolution data by simple linear
            interpolation.
        covar_ext (:obj:`str`, optional):
            Extension in the MaNGA DRP CUBE file to use as the single
            spatial correlation matrix for all wavelength channels.
        on_disk (:obj:`bool`, optional):
            When searching for available files to analyze, search the
            DRP directory path instead of using the data in the DRPall
            file.  
        can_analyze (:obj:`bool`, optional):
            Only construct script files for datacubes that can/should be
            analyzed by the DAP.  See
            :func:`~mangadap.survey.drpcomplete.DRPComplete.can_analyze`.
        log (:obj:`bool`, optional):
            Flag to create the DAP log files.
        dapproc (:obj:`bool`, optional):
            Flag to execute the main DAP processing.
        pltifu_plots (:obj:`bool`, optional):
            Create the QA plate-ifu specific plots.
        post_process (:obj:`bool`, optional):
            Prepare (and possibly perform) the post processing steps:
            create the DAPall file and its associated QA plots, and the
            fit QA plots for each plate.
        post_plots (:obj:`bool`, optional):
            Create the post-processing QA plots.
        report_progress (:obj:`bool`, optional):
            Report the progress of the analysis.
        vebose (:obj:`int`, optional):
            Verbosity level between 0 and 2.
        label (:obj:`str`, optional):
            Label to use in cluster queue.
        nodes (:obj:`int`, optional):
            Number of cluster nodes to use.
        qos (:obj:`str`, optional):
            Select a specific processor (e.g. sdss-fast when the manga
            user).  This option is ignored if :attr:`q` is set.
        umask (:obj:`str`, optional):
            umask to set for output.
        walltime (:obj:`str`, optional):
            Maximum wall time for cluster job.
        hard (:obj:`bool`, optional):
            Same as hard keyword in cluster submission; see
            :func:`pbs.queue.commit`.  This keyword is in the
            **opposite** sense of the "--toughness" command-line
            options; i.e. including --toughness on the command line sets
            hard=False.
        create (:obj:`bool`, optional):
            Use the pbs package to create the scripts.
        submit (:obj:`bool`, optional):
            Submit all jobs to the cluster.
        queue (:obj:`str`, optional):
            Name of the destination queue.  When submitting jobs at
            Utah, this should **not** be set (leaving it at the default
            of None).  When submitting jobs at Portsmouth, this can be
            used to select either sciama1.q, cluster.q (default), or
            sciama3.q.

    Attributes:
        overwrite (bool): **Only used with :attr:`all`**.  Overwrite and
            redo any existing analysis.
        quiet (bool): Suppress output
        print_version (bool): Print the version and then return
        mpl (:class:`mangadap.survey.mangampl.MaNGAMPL`): MPL version;
            see above.
        redux_path (str): The top-level path with the DRP files.
        dapver (str): The DAP version.
        analysis_path (str): The top-level path for the DAP output
            files.
        plan_file (str): Name of the plan file to use for ALL DRP data
            in this run.  Will be None if using the default plan.
        plan (:class:`AnalysisPlanSet`): The set of analysis plans
        daptypes (:obj:`list`): This list of daptypes, one per plan
        platelist (list): List of plates to analyze.
        ifudesignlist (list): List of ifudesigns to analyze.
        combinatorics (bool): Use all unique combinations of the entered
            plate/ifudesign/mode lists to create the full list of DRP
            files to analyze.
        on_disk (bool): See above
        can_analyze (bool): See above
        log (str): See above
        dapproc (bool): Flag to execute the main DAP processing
        pltifu_plots (bool): Create the QA plate-ifu specific plots
        post_process (:obj:`bool`): See above
        post_plots (:obj:`bool`): See above
        report_progress (:obj:`bool`): See abve
        label (str): Label to use in cluster queue.
        nodes (int): Number of cluster nodes to use.
        qos (str): Specific processor to use (Utah specific)
        umask (str): umask to set for output.
        walltime (str): Maximum wall time for cluster job.
        hard (bool): Same as hard keyword in cluster submission; see
            :func:`pbs.queue.commit`.
        create (bool): Use the pbs package to create the submission
            scripts.
        submit (bool): Submit all jobs to the cluster.
        q (str): Queue destination (Portsmouth specific)
        drpc (:class:`mangadap.survey.drpcomplete.DRPComplete`):
            Database of the available DRP files and the parameters
            necessary to write the DAP par files.

    Raises:
        FileNotFoundError:
            Raised if a plate-ifu list file is provided but it does not
            exist, or if the plan file does not exist.
        ValueError:
            Raised if all processing steps are turned off, if the
            selected MPL is undefined; if the fast node is selected with
            nodes > 1; if multiple run modes are selected; or if the
            plate list, ifudesign list, or mode list are not provided
            during a redo mode or if they are provided with any other
            mode.
    """
    def __init__(self, overwrite=None, quiet=False, drpver=None, redux_path=None, dapver=None,
                 analysis_path=None, plan_file=None, platelist=None, ifudesignlist=None,
                 combinatorics=False, list_file=None, sres_ext=None, sres_fill=None,
                 covar_ext=None, on_disk=False, can_analyze=False, log=False, dapproc=True,
                 pltifu_plots=True, post_process=False, post_plots=False, report_progress=False,
                 verbose=0, label='mangadap', nodes=1, cpus=None, qos=None, umask='0027',
                 walltime='240:00:00', hard=True, create=False, submit=False, queue=None):

        # Save run-mode options
        self.overwrite = overwrite
        self.quiet = quiet

        # Override environment
        self.drpver = drpver
        self.redux_path = None if redux_path is None else Path(redux_path).resolve()
        self.dapver = dapver
        self.analysis_path = None if analysis_path is None else Path(analysis_path).resolve()

        # List of files to analyze
        self.plan_file = None if plan_file is None else Path(plan_file).resolve()
        _list_file = None if list_file is None else Path(list_file).resolve()
        if _list_file is not None and not _list_file.exists():
            raise FileNotFoundError(f'No file: {_list_file}')
        if list_file is None:
            self.platelist = arginp_to_list(platelist, evaluate=True)
            self.ifudesignlist = arginp_to_list(ifudesignlist, evaluate=True)
        else:
            if platelist is not None or ifudesignlist is not None:
                warnings.warn('Provided both file and direct list of plate-ifus to process; '
                              'file input takes precedence.')
            self.platelist, self.ifudesignlist = self._read_file_list(_list_file)
        self.combinatorics = combinatorics
        self.sres_ext = sres_ext
        self.sres_fill = sres_fill
        self.covar_ext = covar_ext

        self.on_disk = on_disk
        self.can_analyze = can_analyze

        # Set the options for output
        self.log = log
        self.dapproc = dapproc
        self.pltifu_plots = pltifu_plots
        self.post_process = post_process
        self.post_plots = post_plots

        # Report the progess of the cluster, otherwise just finish
        # script
        self.report_progress = report_progress
        self.verbose = verbose

        # Cluster queue keywords
        self.label = label
        self.nodes = nodes
        self.cpus = cpus
        self.qos = qos
        self.umask = umask
        self.walltime = walltime
        self.hard = hard
        self.create = create
        self.submit = submit
        self.q = queue

        # If submitting to the cluster, make sure that scripts will be
        # created!
        if not self.create and self.submit:
            self.create = True

        # If setting qos, the number of nodes must be 1
        if self.qos is not None and self.nodes > 1:
            raise ValueError('When selecting the fast node, must have nodes=1!')

        # Check processing steps
        nproc = 0
        processes = []
        if self.dapproc:
            nproc += 1
            processes += [ 'Main DAP blocks' ]
        if self.pltifu_plots:
            nproc += 1
            processes += [ 'PLATEIFU-specific QA plots' ]
        if self.post_process:
            nproc += 1
            processes += [ 'DAPall' ]
        if self.post_plots:
            nproc += 1
            processes += [ 'DAPall and PLATE-specific QA plots' ]
        if nproc < 1:
            raise ValueError('No processing steps requested!')

        if (self.dapproc or self.pltifu_plots) and self.submit and self.post_process \
                and not self.report_progress:
            warnings.warn('If both processing and post processing, need to report progress.  '
                          'Setting to report progress.')
            self.report_progress = True

        # Set versions
        if self.drpver is None:
            self.drpver = manga.drp_version()
        if self.dapver is None:
            self.dapver = manga.dap_version()

        # Set the output paths
        self.redux_path = manga.drp_redux_path(self.drpver) \
                                if self.redux_path is None else self.redux_path
        self.analysis_path = manga.dap_analysis_path(self.drpver, self.dapver) \
                                    if self.analysis_path is None else self.analysis_path

        # Set the subdirectory from which the scripts are sourced and
        # the various log files are written.  Currently based on start
        # UTC; e.g., analysis_path/log/19May2016T09.17.42UTC
        self.calling_path = self.analysis_path / 'log' \
                                / time.strftime('%d%b%YT%H.%M.%SUTC',time.gmtime())

        # Create the calling path. The existence test is executed, even
        # though time stamp should mean that calling_path never exists.
        # TODO: allow argument to set calling_path directly?
        if not self.calling_path.exists():
            self.calling_path.mkdir(parents=True)

        # Allow for the default plan.
        if self.plan_file is None:
            warnings.warn('No analysis plan was provided.  DAP executions will adopt the default '
                          'plan.')
        else:
            if not self.plan_file.exists():
                raise FileNotFoundError(f'Plan file does not exist: {self.plan_file}')
            _plan_file = self.plan_file
            self.plan_file = self.calling_path / self.plan_file.name
            if not self.plan_file.exists():
                shutil.copy2(_plan_file, self.plan_file)

        # Construct the DAPTYPEs and check that the plans yield a fully
        # unique list
        self.plan = manga.MaNGAAnalysisPlan.default(analysis_path=self.analysis_path) \
                        if self.plan_file is None \
                        else manga.MaNGAAnalysisPlan.from_toml(self.plan_file,
                                                               analysis_path=self.analysis_path)
        self.daptypes = [self.plan[key]['key'] for key in self.plan.keys()]
        if len(numpy.unique(self.daptypes)) != self.plan.nplans:
            raise ValueError('{0} do not yield unique DAPTYPEs.'.format(
                                'Default plans' if self.plan_file is None 
                                    else f'Plans in {self.plan_file} '))

        # Alert the user of the versions to be used
        if self.q is not None:
            print(f'Attempting to submit to queue: {self.q}')
        print('Versions:')
        for key in python_versions.keys():
            print(f'    {key:>12}: {python_versions[key]:<30}')
        print(f'    {"DRP":>12}: {self.drpver:<30}')
        print(f'    {"DAP":>12}: {self.dapver:<30}')
        print('Paths:')
        print(f'      REDUX: {self.redux_path}')
        print(f'   ANALYSIS: {self.analysis_path}')
        print(f'    CALLING: {self.calling_path}')
        if self.plan_file is not None:
            print(f'  PLAN FILE: {self.plan_file}')
        print('Processing steps added to scripts:')
        print(f'        {processes}')

        # Create and update the DRPComplete file if necessary
        self.drpc = DRPComplete(drpver=self.drpver, redux_path=self.redux_path,
                                dapver=self.dapver, analysis_path=self.analysis_path)
        
        # Update the DRPComplete list; force an update if platetarget
        # files are provided
        self.drpc.update(platelist=self.platelist, ifudesignlist=self.ifudesignlist,
                         combinatorics=self.combinatorics, on_disk=self.on_disk)

        # Hereafter, self.platelist and self.ifuplatelist are the INPUT
        # values, whereas self.drpc.platelist and self.drpc.ifuplatelist
        # are the actual DRP reductions to analyze (including
        # combinatorics and file existence tests).  See, e.g.,
        # selected_drpfile_list().

        # Find the rows in the DRPComplete database with the DRP file
        # meta data
        drpc_rows = self._selected_drpfile_rows(can_analyze=self.can_analyze)

        print(f'Number of DRP files to process: {len(drpc_rows)}')
        if len(drpc_rows) == 0:
            # If coded correctly, this function should never return here.
            return

        # Construct the post-processing scripts before
        # constructing/submitting datacube processing scripts
        if self.post_process or self.post_plots:
            post_queue = self._build_post_queue(drpc_rows)

        # Process the relevant datacubes
        if self.dapproc or self.pltifu_plots:
            # Create the script files for the set of drpfiles, and
            # submit the scripts to the queue if self.submit is true
            proc_queue = self._build_proc_queue(drpc_rows)

            # Submit the queue to the cluster
            if self.create:
                proc_queue.commit(hard=self.hard, submit=self.submit)

            # If not reporting progress, done
            if not self.report_progress:
                return

            # Wait until the queue is finished
            running = True
            while running and self.submit:
                time.sleep(60) 
                percent_complete = proc_queue.get_percent_complete()
                print(f'Percent complete ... {percent_complete:.1f}%', end='\r')
                if percent_complete == 100:
                    running = False
            print(f'Percent complete ... {percent_complete:.1f}%')

        if not self.post_process and not self.post_plots:
            # Don't perform the post-processing
            return

        # Submit the post-processing queue to the cluster
        if self.create:
            post_queue.commit(hard=self.hard, submit=self.submit)

    # ******************************************************************
    #  UTILITY FUNCTIONS
    # ******************************************************************
    def _read_file_list(self, list_file):
        db = numpy.genfromtxt(str(list_file), dtype=object)
        if db.shape[1] != 2:
            raise ValueError('Input file should contain 2 columns, plate and ifu.')
        return db[:,0].astype(int).tolist(), db[:,1].astype(int).tolist()

    def _create_queue(self):
        """
        Create a queue instance.
        
        This function is written to be functional for submission to the
        cluster queues both at Utah and using the sciama cluster at
        Portsmouth.  The available queues are:

        +-----------+--------+------+-----------+
        |     Queue |  Nodes |  PPN |   GB/core |
        +===========+========+======+===========+
        |     ember |     ?? |   12 |        ?? |
        +-----------+--------+------+-----------+
        | sciama1.q |     80 |   12 |         2 |
        +-----------+--------+------+-----------+
        | sciama2.q |     96 |   16 |         4 |
        +-----------+--------+------+-----------+
        | sciama3.q |     48 |   20 |       3.2 |
        +-----------+--------+------+-----------+

        One should **not** submit to sciama3.q without permission from
        Will Percival.  The default queue is cluster.q (sciama2.q).

        Submissions to Utah should have their queue destination set to
        None.  However, the fast node can be selected using :attr:`qos`.
        """
        # Instantiate
        queue = pbs.queue(verbose=not self.quiet)

        if self.q == 'ember':
            # Submitting to Utah ember cluster
            ppn = 12
            cpus = ppn if self.cpus is None else min(self.cpus, ppn)
            walltime = self.walltime if int(self.walltime.split(':')[0]) < 72 else '72:00:00'
            queue.create(label=self.label, nodes=self.nodes, qos=self.qos, umask=self.umask,
                         walltime=walltime, ppn=ppn, cpus=cpus, partition='ember', alloc='sdss')
        elif self.q is not None:
            # All other self.q values expected for Portsmouth cluster,
            # sciama.  In this case, the number of nodes is queue
            # dependent, and qos is not set
            if self.q == 'sciama1.q':
                ppn = 12
            elif self.q == 'sciama3.q':
                ppn = 20
            else:
                ppn = 16
            cpus = ppn if self.cpus is None else min(self.cpus, ppn)
            queue.create(label=self.label, nodes=self.nodes, umask=self.umask,
                         walltime=self.walltime, queue=self.q, ppn=ppn, cpus=cpus)
        else:
            # self.q can be None when submitting to both the Portsmouth
            # and Utah clusters.  In this case, the default queue
            # destination and ppn is correct.  qos is also set, but this
            # should only be used when submitting to Utah.
            ppn = 16
            cpus = ppn if self.cpus is None else min(self.cpus, ppn)
            queue.create(label=self.label, nodes=self.nodes, qos=self.qos, umask=self.umask,
                         walltime=self.walltime, ppn=ppn, cpus=cpus)

        return queue

    # Files:
    #       - mangadap-{plate}-{ifudesign}-LOG{mode} = script file = *
    #       - *.ready = script/directory is ready for queue submission
    #       - *.queued = script submitted to queue
    #       - *.started = script assigned a node/processor and has started running
    #       - *.done = completed execution of the full script
    # Other:
    #       - *.par = parameter file
    #       - *.out, *.err = stdout and stderr output from the script
    def _build_proc_queue(self, drpc_rows):
        """
        Create a queue instance, write the script files for all CPUs
        using :func:`prepare_for_analysis` and append the job to the
        queue, and then submit the jobs to the cluster queue.  If
        :attr:`submit` is False, this is essentially a wrapper for
        :func:`prepare_for_analysis`.

        This function is written to be functional for submission to the
        cluster queues both at Utah and using the sciama cluster at
        Portsmouth; see :func:`_create_queue`.

        .. todo::
            - This algorithm is slow because it has to search through
              the drpcomplete fits file each time to find the
              plate/ifudesign combo.  Should make this more efficient.

        Args:
            drpc_rows (`numpy.ndarray`):
                A vector with the rows in the
                :class:`mangadap.survey.drpcomplete.DRPComplete`
                database with the DRP files to analyze.
        """
        # If submitting to the queue, create the queue object.  The
        # creation of the queue object is system dependent; however,
        # once created, the queue object is treated identically for both
        # the Utah and Portsmouth clusters.
        if self.create:
            queue = self._create_queue()
       
        # Create the script files, regardless of whether or not they are
        # submitted to the queue, appending the scripts to the queue if
        # they are to be submitted
        ndrp = len(drpc_rows)
        for i, index in enumerate(drpc_rows):
            # Write the compute script (write_compute_script also checks
            # the path exists!)
            scriptfile, stdoutfile, stderrfile \
                    = self.write_compute_script(index, dapproc=self.dapproc,
                                                plots=self.pltifu_plots, overwrite=self.overwrite)

            # Set the status to ready
            plate = self.drpc['PLATE'][index]
            ifudesign = self.drpc['IFUDESIGN'][index]
            self.set_pltifu_status(plate, ifudesign, status='ready')

            if self.create:
                queue.append('source {0}'.format(scriptfile), outfile=stdoutfile,
                                  errfile=stderrfile)
                self.set_pltifu_status(plate, ifudesign, status='queued')
            else:
                print(f'Preparing for analysis...{(i+1)*100/ndrp:.1f}%', end='\r')

        if not self.create:
            print(f'Preparing for analysis...100%')

        return queue if self.create else None

    def _build_post_queue(self, drpc_rows):
        """
        Build the queue that performs the post-processing steps.
        """
        # Create the queue
        if self.create:
            queue = self._create_queue()

        if self.post_process:
            # Write the DAPall script and set its status
            dapall_scr, dapall_out, dapall_err \
                    = self.write_dapall_script(plots=self.post_plots, overwrite=self.overwrite)
            self.set_status(dapall_scr, status='ready')
            if self.create:
                # Add it to the queue
                queue.append(f'source {dapall_scr}', outfile=dapall_out, errfile=dapall_err)
                self.set_status(dapall_scr, status='queued')

        # Write the plate QA plot scripts
        if self.post_plots:
            # Find the unique plates
            plates = numpy.unique(self.drpc['PLATE'][drpc_rows])
            for plt in plates:
                # Write the plot script for this plate
                scriptfile, stdoutfile, stderrfile \
                        = self.write_plate_qa_script(plt, overwrite=self.overwrite)
                self.set_status(scriptfile, status='ready')
                
                if self.create:
                    queue.append(f'source {scriptfile}', outfile=stdoutfile, errfile=stderrfile)
                    self.set_status(scriptfile, status='queued')

        return queue if self.create else None

    # ******************************************************************
    # Management
    # ******************************************************************
    def set_pltifu_status(self, plate, ifudesign, status='queued'):
        """
        Generate a touch file that signifies the status for a given
        plate/ifudesign/mode.

        Args:
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                IFU design number
            status (:obj:`str`, optional):
                The status signifier for the touch file.
        """
        # Get the name of the status file        
        root = manga.MaNGAConfig(plate, ifudesign, drpver=self.drpver,
                                 redux_path=self.redux_path).cfg_root
        self.set_status(str(self.calling_path / str(plate) / str(ifudesign) / root), status)
        
    def set_status(self, root, status='queued'):
        """
        Generate a general touch file.

        Args:
            root (:obj:`str`): Full path and root name of the touch file
            status (:obj:`str`, optional):
                The status signifier for the touch file.
        """
        # Touch the status file
        Path(f'{root}.{status}').touch()

    def _selected_drpfile_rows(self, can_analyze=False):
        """
        Find the rows in the
        :class:`mangadap.survey.drpcomplete.DRPComplete` database with
        the selected plates and IFUs to analyze.

        Args:
            can_analyze (:obj:`bool`, optional):
                In addition to returning the selected rows, only include those
                rows with DRP datacubes that can be analyzed.  See
                :func:`mangadap.survey.drpcomplete.DRPComplete.can_analyze`.

        Returns:
            `numpy.ndarray`_: A vector with the rows in the DRPComplete object
            with the relevant DRP file data.
        """
        # Find the rows in the DRPComplete database with the specified
        # plate and IFUS

        # All available PLATEIFUs
        drpc_pltifu = numpy.array([f'{p}-{i}' for p,i in 
                                        zip(self.drpc['PLATE'],self.drpc['IFUDESIGN'])])
        # The PLATEIFUs to analyze
        this_pltifu = numpy.array([f'{p}-{i}' for p,i in 
                                        zip(self.drpc.platelist, self.drpc.ifudesignlist)])

        # The rows with the PLATEIFUs to analyze
        rows = numpy.array([numpy.where(drpc_pltifu == pi)[0][0] if pi in drpc_pltifu else -1 \
                                for pi in this_pltifu])
        if numpy.any(rows < 0):
            raise ValueError('Should be able to find all plates and ifus.')
        if can_analyze:
            keep = self.drpc.can_analyze()[rows]
            rows = rows[keep]
        return rows

    def write_compute_script(self, index, dapproc=True, plots=True, overwrite=False,
                              relative_symlink=True):
        """
        Write the MaNGA DAP script file that is sent to a single CPU to
        analyze a single DRP file with a given plate and ifudesign.

        Args:
            index (:obj:`int`, optional):
                Row in the
                :class:`mangadap.survey.drpcomplete.DRPComplete`
                database with the DRP file to analysis.
            dapproc (:obj:`bool`, optional):
                Flag to execute the main DAP processing.
            plots (:obj:`bool`, optional):
                Create the QA plots.
            overwrite (:obj:`bool`, optional):
                Flag to overwrite any existing files.
            relative_symlink (:obj:`bool`, optional):
                Set the symlink to the par file to be a relative path
                instead of the absolute path.

        Returns:
            :obj:`str`: Three strings with the name of the written
            script file, the file for the output sent to STDOUT, and the
            file for the output sent to STDERR.
        """
        # Fake out the plan cube to just require the configuration information
        self.plan.cube = manga.MaNGAConfig(self.drpc['PLATE'][index],
                                           self.drpc['IFUDESIGN'][index], drpver=self.drpver,
                                           redux_path=self.redux_path)
        cfgfile = self.plan.common_path() / f'{self.plan.cube.cfg_root}.ini'

        # Write the configuration file if it doesn't exist
        if not cfgfile.exists() or overwrite:
            # overwrite defaults to True
            self.drpc.write_config(cfgfile, index=index, sres_ext=self.sres_ext,
                                   sres_fill=self.sres_fill, covar_ext=self.covar_ext)
            # and create symlinks to it
            for i, key in enumerate(self.plan.keys()):
                method_ref_dir = self.plan.method_path(plan_index=i, ref=True)
                create_symlink(str(cfgfile), str(method_ref_dir),
                               relative_symlink=relative_symlink, overwrite=overwrite, quiet=True)

        # Set the root path for the scripts, inputs, outputs, and logs
        _calling_path = self.calling_path / str(self.plan.cube.plate) \
                            / str(self.plan.cube.ifudesign)

        # Set the names for the script, stdout, and stderr files
        scriptfile = _calling_path / f'{self.plan.cube.cfg_root}'
        stdoutfile = Path(f'{scriptfile}.out').resolve()
        stderrfile = Path(f'{scriptfile}.err').resolve()

        ################################################################
        # Script file already exists, so just return
        if scriptfile.exists() and not overwrite:
            return scriptfile, stdoutfile, stderrfile

        # Main script components are:
        #   - (dapproc) Run the DAP processing code
        #   - (plots) Run the plotting scripts to plot the output
        # ONE OF THESE 2 MUST BE TRUE

        ################################################################
        # Open the script file and write the date as a commented header
        # line
        if not scriptfile.parent.exists():
            scriptfile.parent.mkdir(parents=True)

        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write(f'# {time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())}\n')
        file.write('\n')

        # Create the started touch file
        startfile = f'{scriptfile}.started'
        file.write(f'touch {startfile}\n')
        file.write('\n')

        # Command that runs the DAP
        if dapproc:
            command = f'OMP_NUM_THREADS=1 manga_dap -c {cfgfile} -o {self.analysis_path}'
            if self.plan_file is not None:
                command += f' -p {self.plan_file}'
            if self.log:
                command += f' --log {scriptfile}.log'
            if self.verbose > 0:
                command += (' -'+'v'*self.verbose)
            file.write(f'{command}\n')
            file.write('\n')

        # Plotting scripts
        if plots:
            command = f'OMP_NUM_THREADS=1 dap_ppxffit_qa -c {cfgfile} -o {self.analysis_path}'
            if self.plan_file is not None:
                command += f' -p {self.plan_file}'
            command += f' -b 2.5'
            file.write(f'{command}\n')
            file.write('\n')

            command = f'OMP_NUM_THREADS=1 spotcheck_dap_maps -c {cfgfile} -o {self.analysis_path}'
            if self.plan_file is not None:
                command += f' -p {self.plan_file}'
            command += f' -b 2.5'
            file.write(f'{command}\n')
            file.write('\n')

            command = f'OMP_NUM_THREADS=1 dap_fit_residuals -c {cfgfile} -o {self.analysis_path}'
            if self.plan_file is not None:
                command += f' -p {self.plan_file}'
            command += f' -b 2.5'
            file.write(f'{command}\n')
            file.write('\n')

        # Touch the done file
        donefile = f'{scriptfile}.done'
        file.write(f'touch {donefile}\n')
        file.write('\n')

        # Done
        file.close()

        # Return the script file, file for stdout, and file for stderr
        return scriptfile, stdoutfile, stderrfile
    
    def write_dapall_script(self, plots=True, overwrite=False):
        """
        Write the script used to construct the DAPall file and its
        associated quality assessment plots.

        Args:
            plots (:obj:`bool`, optional):
                Create the QA plots. Default is True.
            overwrite (:obj:`bool`, optional):
                Flag to overwrite any existing files.

        Returns:
            :obj:`str`: Three strings with the name of the written
            script file, the file for the output sent to STDOUT, and the
            file for the output sent to STDERR.
        """
        # Check that the path exists, creating it if not
        if not self.calling_path.exists():
            self.calling_path.mkdir(parents=True)
        
        # Set the names for the script, stdout, and stderr files
        scriptfile = self.calling_path / 'build_dapall'
        stdoutfile = f'{scriptfile}.out'
        stderrfile = f'{scriptfile}.err'

        # Script file already exists, so just return
        if scriptfile.exists() and not overwrite:
            return scriptfile, stdoutfile, stderrfile

        # Open the script file and write the date as a commented header
        # line
        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write(f'# {time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())}\n')
        file.write('\n')

        # Create the started touch file
        startfile = f'{scriptfile}.started'
        file.write(f'touch {startfile}\n')
        file.write('\n')

        # Command that constructs the DAPall file
        command = 'OMP_NUM_THREADS=1 '
        command += f'construct_dapall --drpver {self.drpver} -r {self.redux_path} ' \
                   f'--dapver {self.dapver} -a {self.analysis_path}'
        if self.plan_file is not None:
            command += f' --plan_file {self.plan_file}'
        if self.verbose > 0:
            command += (' -'+'v'*self.verbose )
        file.write(f'{command}\n')
        file.write('\n')

        # Add the plotting commands
        if plots:
            command = f'OMP_NUM_THREADS=1 dapall_qa --drpver {self.drpver} ' \
                      f'--redux_path {self.redux_path} --dapver {self.dapver} ' \
                      f'--analysis_path {self.analysis_path}'
            if self.plan_file is not None:
                command += f' --plan_file {self.plan_file}'
            file.write(f'{command}\n')
            file.write('\n')

        # Touch the done file
        donefile = f'{scriptfile}.done'
        file.write(f'touch {donefile}\n')
        file.write('\n')

        file.close()
        ################################################################

        # Return the script file, file for stdout, and file for stderr
        return scriptfile, stdoutfile, stderrfile
    
    def write_plate_qa_script(self, plate, overwrite=False):
        """
        Write the script used to create the plate fit QA plot.

        Args:
            plate (:obj:`int`):
                The plate number for the plot.
            overwrite (:obj:`bool`, optional):
                Flag to overwrite any existing files.

        Returns:
            :obj:`str`: Three strings with the name of the written
            script file, the file for the output sent to STDOUT, and the
            file for the output sent to STDERR.
        """
        # Check that the path exists, creating it if not
        path = self.calling_path / str(plate)
        if not path.exists():
            path.mkdir(parents=True)
        
        # Set the names for the script, stdout, and stderr files
        scriptfile = self.calling_path / str(plate) / f'{plate}_fitqa'
        stdoutfile = f'{scriptfile}.out'
        stderrfile = f'{scriptfile}.err'

        # Script file already exists, so just return
        if scriptfile.exists() and not overwrite:
            return scriptfile, stdoutfile, stderrfile

        # Open the script file and write the date as a commented header
        # line
        file = open(scriptfile, 'w')
        file.write('# Auto-generated batch file\n')
        file.write(f'# {time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())}\n')
        file.write('\n')

        # Create the started touch file
        startfile = f'{scriptfile}.started'
        file.write(f'touch {startfile}\n')
        file.write('\n')

        # Command that creates the plots
        command = f'OMP_NUM_THREADS=1 dap_plate_fit_qa {plate} --analysis_path {self.analysis_path}'
        if self.plan_file is not None:
            command += f' --plan_file {self.plan_file}'
        file.write(f'{command}\n')
        file.write('\n')

        # Touch the done file
        donefile = f'{scriptfile}.done'
        file.write(f'touch {donefile}\n')
        file.write('\n')

        file.close()
        ################################################################

        # Return the script file, file for stdout, and file for stderr
        return scriptfile, stdoutfile, stderrfile
    


